/*
 *******************************************************************
 *** This software is copyright 2008 by Marcel van Tongeren      ***
 *** You have permission to use, modify, copy, and distribute    ***
 *** this software so long as this copyright notice is retained. ***
 *** This software may not be used in commercial applications    ***
 *** without express written permission from the author.         ***
 *******************************************************************
*/

#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif

#ifdef __WXMSW__
#include "wx/msw/private.h"
#endif

#if !wxUSE_COMBOCTRL
    #error "Please set wxUSE_COMBOCTRL to 1 and rebuild the library."
#endif

#include "main.h"
#include "tmc1800.h"

Tmc1800::Tmc1800(const wxString& title, const wxPoint& pos, const wxSize& size, double zoom, double zoomfactor, int computerType)
:Pixie(title, pos, size, zoom, zoomfactor, computerType)
{
}

Tmc1800::~Tmc1800()
{
	p_Main->setMainPos(TMC1800, GetPosition());
}

void Tmc1800::configureComputer()
{
	outType_[1] = VIPKEYOUT;
	outType_[3] = VIPOUT4;
	efType_[2] = VIPEF2;
	efType_[3] = VIPKEYEF;

	p_Main->message("Configuring Telmac TMC-1800");
	p_Main->message("	Output 2: key latch, output 4: address latch");
	p_Main->message("	EF 2: cassette in, EF 3: keyboard\n");

	for (int i=0; i<64; i++)
		hexKeyDefA_[i] = 0;
	if (!p_Main->getUseKeyboard(TMC1800))
	{
		p_Main->getDefaultHexKeys("TMC1800", "A", hexKeyDefA_);
		p_Main->getDefaultGameKeys("TMC1800", "A", keyDefGameValueA_, keyDefGameHexA_);
		p_Main->getDefaultGameKeys("TMC1800", "B", keyDefGameValueB_, keyDefGameHexB_);

		if (p_Main->getConfigBool("/TMC1800/GameAuto", true))
		{
			p_Main->loadKeyDefinition(p_Main->getRamFile(TMC1800), p_Main->getChip8SW(TMC1800), keyDefGameHexA_, keyDefGameHexB_);
			p_Main->storeDefaultGameKeys("TMC1800", "A", keyDefGameValueA_, keyDefGameHexA_);
			p_Main->storeDefaultGameKeys("TMC1800", "B", keyDefGameValueB_, keyDefGameHexB_);
		}
	}
	else
	{
		for (int i=48; i<58; i++)
			hexKeyDefA_[i-48] = i;
		for (int i=65; i<91; i++)
			hexKeyDefA_[i-55] = i;
	}

	resetCpu();
}

void Tmc1800::reDefineKeysA(int hexKeyDefA[], int keyDefGameValueA[], int keyDefGameHexA[])
{
	for (int i=0; i<16; i++)
	{
		hexKeyDefA_[i] = hexKeyDefA[i];
	}
	for (int i=0; i<5; i++)
	{
		keyDefGameValueA_[i] = keyDefGameValueA[i];
		keyDefGameHexA_[i] = keyDefGameHexA[i];
	}
}

void Tmc1800::reDefineKeysB(int* WXUNUSED(hexKeyDefB[]), int keyDefGameValueB[], int keyDefGameHexB[])
{
	for (int i=0; i<5; i++)
	{
		keyDefGameValueB_[i] = keyDefGameValueB[i];
		keyDefGameHexB_[i] = keyDefGameHexB[i];
	}
}

void Tmc1800::initComputer()
{
	setClear(1);
	setWait(1);
	cassetteEf_ = 0;

	telmac1800KeyPort_ = 0;
	for (int i=0; i<64; i++)
		telmac1800KeyState_[i] = 0;

	addressLatch_ = 0x8000;
	runPressed_ = false;
}

void Tmc1800::keyDown(int keycode)
{
	for (int i=0; i<36; i++)
	{
		if (keycode == hexKeyDefA_[i])
		{
			telmac1800KeyState_[i] = 1;
		}
	}
	for (int i=0; i<5; i++)
	{
		if (keycode == keyDefGameValueA_[i])
		{
			telmac1800KeyState_[keyDefGameHexA_[i]&0xf] = 1;
		}
		if (keycode == keyDefGameValueB_[i])
		{
			telmac1800KeyState_[keyDefGameHexB_[i]&0xf] = 1;
		}
	}
}

void Tmc1800::keyUp(int keycode)
{
	for (int i=0; i<36; i++)
	{
		if (keycode == hexKeyDefA_[i])
			telmac1800KeyState_[i] = 0;
	}
	for (int i=0; i<5; i++)
	{
		if (keycode == keyDefGameValueA_[i])
		{
			telmac1800KeyState_[keyDefGameHexA_[i]&0xf] = 0;
		}
		if (keycode == keyDefGameValueB_[i])
		{
			telmac1800KeyState_[keyDefGameHexB_[i]&0xf] = 0;
		}
	}
}

void Tmc1800::onRun()
{
	runPressed_ = true;
}

Byte Tmc1800::ef(int flag)
{
	switch(efType_[flag])
	{
		case 0:
			return 1;
		break;

		case PIXIEEF:
			return efPixie();
		break;

		case VIPEF2:
			return cassetteEf_;
		break;

		case VIPKEYEF:
			return ef3();
		break;

		default:
			return 1;
	}
}

Byte Tmc1800::ef3()
{
	return (telmac1800KeyState_[telmac1800KeyPort_]) ? 0 : 1;
}

Byte Tmc1800::in(Byte port, Word WXUNUSED(address))
{
	Byte ret;

	switch(inType_[port-1])
	{
		case 0:
			ret = 255;
		break;

		case PIXIEIN:
			ret = inPixie();
		break;

		case PIXIEOUT:
			outPixie();
			return 0;
		break;

		default:
			ret = 255;
	}
	inValues_[port] = ret;
	return ret;
}

void Tmc1800::out(Byte port, Word WXUNUSED(address), Byte value)
{
	outValues_[port] = value;

	switch(outType_[port-1])
	{
		case 0:
			return;
		break;

		case PIXIEOUT:
			outPixie();
		break;

		case VIPKEYOUT:
			outTMC1800(value);
		break;

		case VIPOUT4:
			addressLatch_ = 0;
		break;
	}
}

void Tmc1800::outTMC1800(Byte value)
{
	telmac1800KeyPort_ = value&0x3f;
}

void Tmc1800::cycle(int type)
{
	switch(cycleType_[type])
	{
		case 0:
			return;
		break;

		case PIXIECYCLE:
			cyclePixie();
		break;
	}
}

void Tmc1800::startComputer()
{
	resetPressed_ = false;

	p_Main->setSwName("");
	p_Main->updateTitle();

	readProgram(p_Main->getRomDir(TMC1800, MAINROM1), p_Main->getRomFile(TMC1800, MAINROM1), ROM, 0x8000, NONAME);

	switch(p_Main->getRamType(TMC1800))
	{
		case 0:
			defineMemoryType(0x0, 0xfff, RAM);
			ramMask_ = 0xfff;
		break;

		case 1:
			defineMemoryType(0x0, 0x3fff, RAM);
			ramMask_ = 0x3fff;
		break;

		case 2:
			defineMemoryType(0x0, 0x7fff, RAM);
			ramMask_ = 0x7fff;
		break;
	}

	readProgram(p_Main->getRamDir(TMC1800), p_Main->getRamFile(TMC1800), NOCHANGE, 0, SHOWNAME);
	readProgram(p_Main->getChip8Dir(TMC1800), p_Main->getChip8SW(TMC1800), NOCHANGE, 0x200, SHOWNAME);
	if (mainMemory_[0x100] ==  0 && mainMemory_[0x1b] == 0x96 && mainMemory_[0x1c] == 0xb7)
		chip8type_ = CHIP8;

	double zoom = p_Main->getZoom();

	configurePixieTmc1800();
	initPixie();
	setZoom(zoom);
	Show(true);

	p_Main->updateTitle();


	cpuCycles_ = 0;
	p_Main->startTime();

	threadPointer->Run();
}

Byte Tmc1800::readMem(Word addr)
{
	if (addr < 0x8000)
		address_ = (addr | addressLatch_) & (ramMask_ | 0x8000);
	else
		address_ = addr & 0x81ff;

	if (memoryType_[address_/256] == UNDEFINED) return 255;
	return mainMemory_[address_| addressLatch_];
}

void Tmc1800::writeMem(Word addr, Byte value, bool writeRom)
{
	if (!writeRom)
		address_ = addr & ramMask_;
	else
		address_ = addr;

	if (mainMemory_[address_]==value)
		return;
	if (!writeRom)
		if (memoryType_[address_/256] != RAM)  return;

	mainMemory_[address_]=value;
	if (writeRom)
		return;
	if (address_ >= (memoryStart_& ramMask_) && address_<((memoryStart_& ramMask_) + 256))
		p_Main->updateDebugMemory(address_);
}

void Tmc1800::cpuInstruction()
{
	if (cpuMode_ == RUN)
	{
		if (steps_ != 0)
		{
			cycle0_=0;
			machineCycle();
			if (cycle0_ == 0) machineCycle();
			if (cycle0_ == 0 && steps_ != 0)
			{
				cpuCycle();
				cpuCycles_ += 2;
			}
			if (debugMode_)
				p_Main->showInstructionTrace();
		}
		else
			soundCycle();

		playSaveLoad();
		checkTMC1800Function();

		if (resetPressed_)
		{
			resetCpu();
			resetPressed_ = false;
			addressLatch_ = 0x8000;
			initPixie();
		}
		if (runPressed_)
		{
			setClear(0);
			p_Main->updateTitle();
			runPressed_ = false;
		}
		if (debugMode_)
			p_Main->cycleDebug();
		p_Main->cycleChip8Debug();
	}
	else
	{
		if (runPressed_)
		{
			setClear(1);
			addressLatch_ = 0x8000;
			initPixie();
			p_Main->updateTitle();
			runPressed_ = false;
		}
	}
}

void Tmc1800::onReset()
{
	resetPressed_ = true;
}

void Tmc1800::checkTMC1800Function()
{
	switch(scratchpadRegister_[programCounter_])
	{
		case 0x80c0:
			p_Main->stopCassette();
		break;

		case 0x80ed:
			p_Main->stopCassette();
		break;

		case 0x8091:	// SAVE
			p_Main->startCassetteSave();
		break;

		case 0x80c2:	// LOAD
			p_Main->startCassetteLoad();
		break;
	}
}
